/*
 * Copyright (c) 2011-2014 Robert Scheck <zarafa@robert-scheck.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * ALTERNATIVELY, this product may be distributed under the terms of
 * the GNU Public License, in which case the provisions of the GPL are
 * required INSTEAD OF the above restrictions.  (This clause is
 * necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <iostream>
#include <zarafa/CommonUtil.h>
#include <mapi4linux/mapidefs.h>
#include <mapi4linux/mapix.h>

// Remove Zarafa leak
#ifdef LOG_DEBUG
# undef LOG_DEBUG
#endif

#include "config.h"

#ifdef STDC_HEADERS
# include <stdio.h>
#endif

#ifdef HAVE_STRING_H
# include <string.h>
#endif

#ifdef HAVE_SYSLOG_H
# include <syslog.h>
#endif

#ifdef HAVE_LIMITS_H
# include <limits.h>
#endif

#ifdef UNICODE
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
# ifdef HAVE_LOCALE_H
#  include <locale.h>
# endif
#endif

#ifdef HAVE_ZARAFA_ECFEATURES_H
# include <zarafa/ECFeatures.h>
#endif

#ifdef HAVE_LIBINTL_H
# include <libintl.h>
# define _(msgid) dgettext("Linux-PAM", msgid)
#else
# define _(msgid) (msgid)
#endif

// Indicate that the following groups are defined
#define PAM_SM_AUTH
#define PAM_SM_ACCOUNT

#if defined(HAVE_SECURITY_PAM_MODULES_H)
# include <security/pam_modules.h>
#elif defined(HAVE_PAM_PAM_MODULES_H)
# include <pam/pam_modules.h>
#endif

#if defined(HAVE_SECURITY_PAM_EXT_H)
# include <security/pam_ext.h>
#elif defined(HAVE_PAM_PAM_EXT_H)
# include <pam/pam_ext.h>
#endif

#if defined(HAVE_SECURITY__PAM_MACROS_H)
# include <security/_pam_macros.h>
#elif defined(HAVE_PAM__PAM_MACROS_H)
# include <pam/_pam_macros.h>
#endif

#if defined(HAVE_SYSLOG_H) && !defined(HAVE_PAM_SYSLOG)
#ifndef LOG_AUTHPRIV
# define LOG_AUTHPRIV LOG_AUTH
#endif 
static void pam_syslog(const pam_handle_t *pamh, int priority, const char *fmt, ...)
{
  va_list args;

  va_start(args, fmt);
  openlog(PACKAGE_NAME, LOG_CONS|LOG_PID, LOG_AUTHPRIV);
  vsyslog(priority, fmt, args);
  va_end(args);
  closelog();
}
#endif

#if !defined(HAVE_PAM_PROMPT)
int pam_vprompt(pam_handle_t *pamh, int style, char **response, const char *fmt, va_list args)
{
  struct pam_message msg;
  struct pam_response *pam_resp = NULL;
  const struct pam_message *pmsg;
  const struct pam_conv *conv;
  const void *convp;
  char *msgbuf;
  int ret;

  if(response)
  {
    *response = NULL;
  }

  ret = pam_get_item(pamh, PAM_CONV, &convp);

  if(ret != PAM_SUCCESS)
  {
    return ret;
  }

  conv = (const struct pam_conv *) convp;

  if(conv == NULL || conv->conv == NULL)
  {
    pam_syslog(pamh, LOG_ERR, "no conversation function");
    return PAM_SYSTEM_ERR;
  }

  if(vasprintf(&msgbuf, fmt, args) < 0)
  {
    pam_syslog(pamh, LOG_ERR, "vasprintf: %m");
    return PAM_BUF_ERR;
  }

  msg.msg_style = style;
  msg.msg = msgbuf;
  pmsg = &msg;

  ret = conv->conv(1, &pmsg, &pam_resp, conv->appdata_ptr);

  if(ret != PAM_SUCCESS && pam_resp != NULL)
  {
    pam_syslog(pamh, LOG_WARNING, "unexpected response from failed conversation function");
  }

  if(response)
  {
    *response = pam_resp == NULL ? NULL : pam_resp->resp;
  }
  else if(pam_resp && pam_resp->resp)
  {
    _pam_overwrite(pam_resp->resp);
    _pam_drop(pam_resp->resp);
  }

  _pam_overwrite(msgbuf);
  _pam_drop(pam_resp);
  free(msgbuf);

  if(ret != PAM_SUCCESS)
  {
    pam_syslog(pamh, LOG_ERR, "conversation failed");
  }

  return ret;
}

int pam_prompt(pam_handle_t *pamh, int style, char **response, const char *fmt, ...)
{
  va_list args;
  int ret;

  va_start(args, fmt);
  ret = pam_vprompt(pamh, style, response, fmt, args);
  va_end(args);

  return ret;
}
#endif

// Conversation function to obtain the user's password
static int pam_get_authtok_external(pam_handle_t *pamh)
{
  char *resp = NULL;
  const void *authtok;
  int ret;

  ret = pam_prompt(pamh, PAM_PROMPT_ECHO_OFF, &resp, _("Password: "));

  if(ret != PAM_SUCCESS || resp == NULL)
  {
    return PAM_AUTHTOK_ERR;
  }

  // Set the auth token
  ret = pam_set_item(pamh, PAM_AUTHTOK, resp);

  // Clean it up
  _pam_overwrite(resp);
  _pam_drop(resp);

  if(ret != PAM_SUCCESS)
  {
    return ret;
  }

  return pam_get_item(pamh, PAM_AUTHTOK, &authtok);
}

// User authentication
PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
  const char *username, *socket = "http://localhost:236/zarafa";
  const void *password;
#ifdef HAVE_ZARAFA_ECFEATURES_H
  char *service, *services = NULL, disabled[CHAR_MAX] = { '\0' };
  IMsgStore *lpStore = NULL;
  IAddrBook *lpAddrBook = NULL;
  int pos;
#endif
  int ret = PAM_AUTH_ERR, debug = 0, use_first_pass = 0, try_first_pass = 0, quiet = 0;
  HRESULT hr = hrSuccess;
  LPMAPISESSION lpSession = NULL;

  // Walk through arguments
  for(; argc-- > 0; ++argv)
  {
    if(!strcmp(*argv, "debug"))
    {
      debug = 1;
    }
    else if(!strcmp(*argv, "quiet"))
    {
      quiet = 1;
    }
    else if(!strcmp(*argv, "use_first_pass"))
    {
      use_first_pass = 1;
    }
    else if(!strcmp(*argv, "try_first_pass"))
    {
      try_first_pass = 1;
    }
    else if(!strncmp(*argv, "socket=", 7))
    {
      if(*(*argv + 7) == '\0')
      {
        pam_syslog(pamh, LOG_ERR, "missing path for socket, using default");
      }
      else
      {
        socket = *argv + 7;
      }
    }
    else if(!strncmp(*argv, "service=", 8))
    {
#ifdef HAVE_ZARAFA_ECFEATURES_H
      if(*(*argv + 8) == '\0')
      {
        pam_syslog(pamh, LOG_ERR, "missing value(s) for service, ignoring");
      }
      else
      {
        services = (char *) *argv + 8;
      }
#else
      pam_syslog(pamh, LOG_ERR, "service option was disabled during compile time, ignoring");
#endif
    }
    else
    {
      pam_syslog(pamh, LOG_ERR, "unknown option %s", *argv);
    }
  }

  // Argument related debugging needs to happen after argument assignment
  if(debug == 1)
  {
    if(strcmp(socket, "http://localhost:236/zarafa"))
    {
      pam_syslog(pamh, LOG_INFO, "using new socket path %s", socket);
    }

    if(services != NULL)
    {
      pam_syslog(pamh, LOG_INFO, "using service%s %s", strstr(services, "|") != NULL ? "s" : "", services);
    }
  }

  // Get the username
  ret = pam_get_user(pamh, &username, NULL);

  if(ret != PAM_SUCCESS || username == NULL)
  {
    pam_syslog(pamh, LOG_ERR, "can not get the username");
    return PAM_SERVICE_ERR;
  }

  if(use_first_pass == 0 && try_first_pass == 0)
  {
    // Converse to obtain a password
    ret = pam_get_authtok_external(pamh);

    if(ret != PAM_SUCCESS)
    {
      pam_syslog(pamh, LOG_ERR, "can not obtain password from user");
      return ret;
    }
  }

  // Check if we got a password
  ret = pam_get_item(pamh, PAM_AUTHTOK, &password);

  if(ret != PAM_SUCCESS || password == NULL)
  {
    if(try_first_pass == 1)
    {
      // Converse to obtain a password
      ret = pam_get_authtok_external(pamh);

      if(ret != PAM_SUCCESS)
      {
        pam_syslog(pamh, LOG_ERR, "can not obtain password from user");
        return ret;
      }

      ret = pam_get_item(pamh, PAM_AUTHTOK, &password);
    }

    if(ret != PAM_SUCCESS || password == NULL)
    {
      pam_syslog(pamh, LOG_ERR, "can not recover user password");
      return PAM_AUTHTOK_RECOVERY_ERR;
    }
  }

  if(debug == 1)
  {
    pam_syslog(pamh, LOG_INFO, "verifying user %s with a password", username);
  }

  // MAPI initialization
  hr = MAPIInitialize(NULL);

  if(hr != hrSuccess)
  {
    pam_syslog(pamh, LOG_ERR, "unable to initialize MAPI session");
    return PAM_SERVICE_ERR;
  }

// Zarafa 7.0.0 and later provides unicode support
#ifdef UNICODE
  // Behavior of mbstowcs() depends on LC_CTYPE
  if(setlocale(LC_CTYPE, "en_US.UTF-8") == NULL)
  {
    pam_syslog(pamh, LOG_ERR, "setlocale() for mbstowcs() failed");
    return PAM_SERVICE_ERR;
  }

  // Convert username from char* to wchar_t* and avoid a buffer overflow, see:
  // https://buildsecurityin.us-cert.gov/bsi/articles/knowledge/coding/769-BSI.html
  unsigned int username_s = mbstowcs(NULL, username, 0) + 1;

  if(username_s == 0 || username_s > ULONG_MAX / sizeof(wchar_t))
  {
    pam_syslog(pamh, LOG_ERR, "mbstowcs() for user %s failed", username);
    return PAM_SERVICE_ERR;
  }

  wchar_t *username_w = (wchar_t *) malloc(username_s * sizeof(wchar_t));
  mbstowcs(username_w, username, username_s);

  // Convert password from char* to wchar_t* and avoid a buffer overflow, see:
  // https://buildsecurityin.us-cert.gov/bsi/articles/knowledge/coding/769-BSI.html
  unsigned int password_s = mbstowcs(NULL, (const char *) password, 0) + 1;

  if(password_s == 0 || password_s > ULONG_MAX / sizeof(wchar_t))
  {
    if(debug == 1)
    {
      pam_syslog(pamh, LOG_ERR, "mbstowcs() for password %s failed", (const char *) password);
    }
    else
    {
      pam_syslog(pamh, LOG_ERR, "mbstowcs() for the password failed");
    }

    return PAM_SERVICE_ERR;
  }

  wchar_t *password_w = (wchar_t *) malloc(password_s * sizeof(wchar_t));
  mbstowcs(password_w, (const char *) password, password_s);

  // Try MAPI login
  hr = HrOpenECSession(&lpSession, username_w, password_w, socket, 0, NULL, NULL);
#else
  // Try MAPI login
  hr = HrOpenECSession(&lpSession, (char *) username, (char *) password, (char *) socket, 0, NULL, NULL);
#endif

#ifdef HAVE_ZARAFA_ECFEATURES_H
  if(services != NULL && hr == hrSuccess)
  {
    // Try to open default store in MAPI session
    hr = HrOpenDefaultStore(lpSession, &lpStore);

    if(hr != hrSuccess)
    {
      pam_syslog(pamh, LOG_ERR, "failed to open MAPI default store");
      ret = PAM_SERVICE_ERR;
    }
    else
    {
      // Try to open address book in MAPI session
      hr = lpSession->OpenAddressBook(0, NULL, AB_NO_DIALOG, &lpAddrBook);

      if(hr != hrSuccess)
      {
        pam_syslog(pamh, LOG_ERR, "failed to open MAPI address book");
        ret = PAM_SERVICE_ERR;
      }
      else
      {
        // Walk through given services and succeed if one feature is enabled
        hr = MAPI_E_LOGON_FAILED;
        pos = 0;

        while((service = strsep(&services, "|")) != NULL)
        {
          // Check if it is a feature because non-existing features are enabled
          if(!isFeature(service) || isFeatureDisabled(service, lpAddrBook, lpStore))
          {
            if(disabled[0] != '\0')
            {
              pos += snprintf(disabled + pos, CHAR_MAX - pos, "|");
            }

            pos += snprintf(disabled + pos, CHAR_MAX - pos, "%s", service);
          }
          else
          {
            hr = hrSuccess;
          }
        }

        if(hr != hrSuccess && (debug == 1 || quiet == 0))
        {
          pam_syslog(pamh, LOG_INFO, "service%s %s not enabled in MAPI for user %s", strstr(disabled, "|") ? "s" : "", disabled, username);
        }
      }
    }
  }
#endif

  // Clean up afterwards
#ifdef HAVE_ZARAFA_ECFEATURES_H
  if(lpAddrBook)
  {
    lpAddrBook->Release();
  }

  if(lpStore)
  {
    lpStore->Release();
  }
#endif

  if(lpSession)
  {
    lpSession->Release();
  }

  MAPIUninitialize();

  // Evaluate the MAPI login
  if(hr == hrSuccess)
  {
    if(debug == 1)
    {
      pam_syslog(pamh, LOG_NOTICE, "authentication succeeded for user %s", username);
    }

    if(pam_set_data(pamh, PACKAGE_NAME, (void *) username, NULL) != PAM_SUCCESS)
    {
      pam_syslog(pamh, LOG_ERR, "pam_set_data() for user %s failed", username);
      return PAM_SERVICE_ERR;  
    }

    return PAM_SUCCESS;
  }
  else if(hr == MAPI_E_NETWORK_ERROR)
  {
    pam_syslog(pamh, LOG_ERR, "unable to connect to server for user %s: 0x%X (MAPI_E_NETWORK_ERROR)", username, hr);
    return PAM_AUTHINFO_UNAVAIL;
  }
  else if(hr == MAPI_E_LOGON_FAILED)
  {
    if(debug == 1 || quiet == 0)
    {
      pam_syslog(pamh, LOG_ERR, "authentication failed for user %s: 0x%X (MAPI_E_LOGON_FAILED)", username, hr);
    }

    return PAM_AUTH_ERR;
  }
#ifdef HAVE_ZARAFA_ECFEATURES_H
  else if(ret == PAM_SERVICE_ERR)
  {
    return ret;
  }
#endif
  else
  {
    pam_syslog(pamh, LOG_ERR, "authentication failed for user %s: 0x%X", username, hr);
    return PAM_AUTH_ERR;
  }

  // Should not be reached
  return PAM_SILENT;
}

// Alter credentials
PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
  return PAM_SUCCESS;
}

// Account management
PAM_EXTERN int pam_sm_acct_mgmt(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
  const void *username;
  int debug = 0;

  // Walk through arguments
  for(; argc-- > 0; ++argv)
  {
    if(!strcmp(*argv, "debug"))
    {
      debug = 1;
    }
  }

  // Check for successful authentication
  if(pam_get_data(pamh, PACKAGE_NAME, &username) != PAM_SUCCESS)
  {
    if(debug == 1)
    {
      pam_syslog(pamh, LOG_INFO, "pam_get_data() did not find any module data");
    }

    // Valid account maybe exists
    return PAM_IGNORE;
  }

  if(debug == 1)
  {
    pam_syslog(pamh, LOG_INFO, "pam_get_data() returned entry %s", (const char *) username);
  }

  return PAM_SUCCESS;
}

// For the case that this module is static
#ifdef PAM_STATIC
struct pam_module _pam_mapi_modstruct = {
  PACKAGE_NAME,
  pam_sm_authenticate,
  pam_sm_setcred,
  pam_sm_acct_mgmt,
  NULL,
  NULL,
  NULL
};
#endif
